#![warn(clippy::pedantic)]
#![no_std]
#![no_main]

use cyw43::Control;
use cyw43_pio::{DEFAULT_CLOCK_DIVIDER, PioSpi};
use defmt::{info, unwrap};
use embassy_executor::Spawner;
use embassy_futures::select::{Either, select};
use embassy_rp::bind_interrupts;
use embassy_rp::config::Config;
use embassy_rp::gpio::{self, Input};
use embassy_rp::peripherals::{DMA_CH0, PIO0};
use embassy_rp::pio::{InterruptHandler, Pio};
use embassy_sync::blocking_mutex::raw::CriticalSectionRawMutex;
use embassy_sync::channel::{Channel, Sender};
use embassy_time::{Duration, Instant, Timer};
use gpio::{Level, Output};
use static_cell::StaticCell;
use {defmt_rtt as _, panic_probe as _};

const ONE_MINUTE_SECONDS: u64 = 60;
const ONE_HOUR_SECONDS: u64 = ONE_MINUTE_SECONDS * 60;
const ONE_DAY_SECONDS: u64 = ONE_HOUR_SECONDS * 24;

const FRESH_TO_STALE_DURATION: Duration = Duration::from_secs(ONE_DAY_SECONDS);
const STALE_TO_OLD_DURATION: Duration = Duration::from_secs(ONE_DAY_SECONDS);

bind_interrupts!(struct Irqs {
    PIO0_IRQ_0 => InterruptHandler<PIO0>;
});

#[embassy_executor::task]
async fn cyw43_task(
    runner: cyw43::Runner<'static, Output<'static>, PioSpi<'static, PIO0, 0, DMA_CH0>>,
) -> ! {
    runner.run().await
}

#[derive(Clone, Copy)]
enum WaterState {
    Fresh,
    Stale,
    Old,
}

impl WaterState {
    fn decay(&mut self) {
        match self {
            Self::Fresh => *self = Self::Stale,
            Self::Stale => *self = Self::Old,
            Self::Old => (),
        }
    }

    fn refill(&mut self) {
        *self = Self::Fresh;
    }

    fn timer(self) -> Timer {
        match self {
            WaterState::Fresh => Timer::after(FRESH_TO_STALE_DURATION),
            WaterState::Stale => Timer::after(STALE_TO_OLD_DURATION),
            // I hope oolong has been watered after this long...
            WaterState::Old => Timer::at(Instant::MAX),
        }
    }
}

#[derive(Clone, Copy)]
enum ChannelMessage {
    Refilled,
}
static CHANNEL: Channel<CriticalSectionRawMutex, ChannelMessage, 8> = Channel::new();

#[embassy_executor::task]
async fn button(
    mut input: Input<'static>,
    sender: Sender<'static, CriticalSectionRawMutex, ChannelMessage, 8>,
) {
    loop {
        info!("Waiting for falling edge");
        input.wait_for_falling_edge().await;
        sender.send(ChannelMessage::Refilled).await;
        input.wait_for_high().await;
    }
}

#[embassy_executor::task]
async fn heartbeat(mut control: Control<'static>) {
    loop {
        info!("Heartbeating!");
        control.gpio_set(0, true).await;
        Timer::after(Duration::from_millis(1)).await;
        control.gpio_set(0, false).await;
        Timer::after(Duration::from_secs(10)).await;
    }
}

#[embassy_executor::main]
async fn main(spawner: Spawner) {
    // let p = embassy_rp::init(Default::default());
    // let mut led = Output::new(p.PIN_18, Level::Low);

    // loop {
    //     info!("led on!");
    //     led.set_high();
    //     Timer::after_secs(1).await;

    //     info!("led off!");
    //     led.set_low();
    //     Timer::after_secs(1).await;
    // }

    let p = embassy_rp::init(Config::default());
    let fw = include_bytes!("../embassy/cyw43-firmware/43439A0.bin");
    let clm = include_bytes!("../embassy/cyw43-firmware/43439A0_clm.bin");

    // To make flashing faster for development, you may want to flash the firmwares independently
    // at hardcoded addresses, instead of baking them into the program with `include_bytes!`:
    //     probe-rs download ../../cyw43-firmware/43439A0.bin --binary-format bin --chip RP2040 --base-address 0x10100000
    //     probe-rs download ../../cyw43-firmware/43439A0_clm.bin --binary-format bin --chip RP2040 --base-address 0x10140000
    //let fw = unsafe { core::slice::from_raw_parts(0x10100000 as *const u8, 230321) };
    //let clm = unsafe { core::slice::from_raw_parts(0x10140000 as *const u8, 4752) };

    let pwr = Output::new(p.PIN_23, Level::Low);
    let cs = Output::new(p.PIN_25, Level::High);
    let mut pio = Pio::new(p.PIO0, Irqs);
    let spi = PioSpi::new(
        &mut pio.common,
        pio.sm0,
        DEFAULT_CLOCK_DIVIDER,
        pio.irq0,
        cs,
        p.PIN_24,
        p.PIN_29,
        p.DMA_CH0,
    );

    let (_net_device, mut control, runner) = {
        static STATE: StaticCell<cyw43::State> = StaticCell::new();
        let state = STATE.init(cyw43::State::new());
        cyw43::new(state, pwr, spi, fw).await
    };

    unwrap!(spawner.spawn(cyw43_task(runner)));

    control.init(clm).await;
    control
        .set_power_management(cyw43::PowerManagementMode::Aggressive)
        .await;
    unwrap!(spawner.spawn(heartbeat(control)));

    unwrap!(spawner.spawn(button(
        {
            let input = Input::new(p.PIN_15, gpio::Pull::Up);
            input
        },
        CHANNEL.sender()
    )));

    let recv = CHANNEL.receiver();

    let mut water_state = WaterState::Fresh;
    let mut timer = water_state.timer();
    let mut red_led = Output::new(p.PIN_13, Level::Low);
    let mut yellow_led = Output::new(p.PIN_18, Level::Low);

    loop {
        match select(&mut timer, recv.receive()).await {
            Either::First(()) => {
                info!("Timer decay received");
                water_state.decay();
                set_leds(water_state, &mut yellow_led, &mut red_led);
                timer = water_state.timer();
            }
            Either::Second(ChannelMessage::Refilled) => {
                info!("Water refilled received");
                water_state.refill();
                set_leds(water_state, &mut yellow_led, &mut red_led);
                timer = water_state.timer();
            }
        }
    }
}

fn set_leds(water_state: WaterState, yellow_led: &mut Output, red_led: &mut Output) {
    match water_state {
        WaterState::Fresh => {
            yellow_led.set_low();
            red_led.set_low();
        }
        WaterState::Stale => {
            yellow_led.set_high();
            red_led.set_low();
        }
        WaterState::Old => {
            yellow_led.set_low();
            red_led.set_high();
        }
    }
}
